#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <sys/times.h>
#include <sys/types.h>

#include "mc_bench.h"

#ifdef SGI64
#define SGI64V2
#endif

void init_rnd(void);
int   rnd(unsigned int *);
double rndf(unsigned int *);
double rndKO(double);
double rndKOS(double);
double gauss(double);

int mc_run(int);
void init_structure(void);
double diff_energy(int);
double total_energy(void);
double proof_energy(double);
double diff_mix(double,double);
double diff_miz(double,double);

double px[PANZ],py[PANZ],pz[PANZ],ptx,pty,ptz;
double etab[PANZ][PANZ];


unsigned int seed,os;

FILE *fpl,*fpt;



/******/
/*MAIN*/
/******/
int main(int argc, char *argv[])
{
 double runTime,runSecs,runMin;
 int trun=-1;
 char file[256];
 struct tms start,stop;


 trun = atoi(argv[1]);

 sprintf(file,"result.%s",argv[2]);

 if(trun<=0)
  {
   printf("MC-Bench needs TIME_STEPS to be specified in mcb.conf\n");
   printf("Terminating\n");
   return(-1);
  }
  
 printf("running MCBench with %d particles for %d steps\n",PANZ,trun);
 init_rnd();
 seed =1234567l;
 fpl = fopen("mcb.log","w");
 if(fpl == NULL)
   printf("Can't open outputfile\n");
 init_structure();

#ifdef WITH_PRINTF
 printf("Starting  MC-Bench\n");
#endif

 times(&start);
 mc_run(trun);
 times(&stop);

 runTime = ((stop.tms_stime +stop.tms_utime) - (start.tms_stime +start.tms_utime)) / (double)sysconf(_SC_CLK_TCK);
 runSecs = fmod(runTime ,60.);
 runMin  = runTime-runSecs;
 fpt=fopen(file,"w");
 printf("Laufzeit %f:%f\n",runMin,runSecs);
 fprintf(fpt,"%d %f %d\n",trun,runTime,PANZ);
 fclose(fpl);
 fclose(fpt);

 return(0);
}

/***************************************************/
/* Initialisierung der Zufallsgenerator Wortl"ange */
/***************************************************/
void init_rnd() 
{ 
 os   = (unsigned long) 256*256*256*128; 
} 

/***************************************/
/* Erzeugung von integer Zufallszahlen */
/***************************************/
int rnd(unsigned int *seed) 
{ 
 unsigned int j; 
 
 j     = (69069l* *seed+1) ; 
 *seed = j; 
 return(j); 
} 

/***************************************/
/* Erzeugung von double  Zufallszahlen */
/***************************************/
double rndf(unsigned int *seed) 
{ 
 unsigned int j; 
 double r; 
  
 j     = (69069l* *seed+1); 
 *seed = j; 
 r     = (double)j/os*.5; 
 return(r); 
} 
/**********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [0,L[ */
/**********************************************************/
double rndKO(double L)
{ 
 return(L*rndf(&seed));
}

/***********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [-L,L[ */
/***********************************************************/
double rndKOS(double L)
{ 
 return(L*(2*rndf(&seed)-1));
}

/***********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [-L,L[ */
/*           mit angen"aherter Gau"sverteilung             */
/***********************************************************/
double gauss(double L)
{
 return(L*((rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)-4)/4));
}



void init_structure()
{
 int p1;
 for(p1=0 ;p1<PANZ; p1++)
  {
   px[p1] = rndKO(WIDTH);
   py[p1] = rndKO(WIDTH);
   pz[p1] = rndKO(HEIGHT);
  }
 /* printf( "%f %f %f\n",px[PANZ/2],py[PANZ/2],pz[PANZ/2]);
 printf( "%f %f %f\n",px[PANZ/2+1],py[PANZ/2+1],pz[PANZ/2+1]);*/
}

int mc_run(int trun)
{
 double e,de,mc_step =0.1,w,r,ea;
 int pn,p2,t=-1,ac=0,gc=0;

 e = total_energy();
 fprintf(fpl,"%6d) E=%12g\n",t,e);
 printf("Start : E=%12g\n",e);
 for(t=0 ; t<trun; t++)
  {
   for(pn=0 ; pn<PANZ; pn++)
    {
     ptx = px[pn]+ gauss(mc_step);
     pty = py[pn]+ gauss(mc_step);
     ptz = pz[pn]+ gauss(mc_step);
     if(ptx>WIDTH)
       ptx-=WIDTH;
     if(pty>WIDTH)
       pty-=WIDTH;
     if(ptz>HEIGHT)
       ptz-=HEIGHT;
     if(ptx<0)
       ptx+= WIDTH;
     if(pty<0)
       pty+= WIDTH;
     if(ptz<0)
       ptz+=HEIGHT;
     gc++;
     de = diff_energy(pn);
     if(de<0)
      {
       e+=de;
       px[pn] =ptx;
       py[pn] =pty;
       pz[pn] =ptz;
       for(p2=0 ; p2<pn; p2++)	 
		 etab[p2][pn] = etab[pn][p2];
       for(p2=pn+1 ; p2<PANZ; p2++)	 
		 etab[pn][p2] = etab[p2][pn];
       ac++;
      }
     else
      {
       w = exp( -de/AKT_TEMP);
       r=rndf(&seed);
       if(w > r)
		{
		 e+=de;
		 px[pn] =ptx;
		 py[pn] =pty;
		 pz[pn] =ptz;
		 for(p2=0 ; p2<pn; p2++)	 
		   etab[p2][pn] = etab[pn][p2];
		 for(p2=pn+1 ; p2<PANZ; p2++)	 
		   etab[pn][p2] = etab[p2][pn];
		 ac++;
		}
      }
    }
/*   ea= proof_energy(e);*/
   if( (t % TNEW) == 0)
    {
     ea=e;
     e = total_energy();
     fprintf(fpl,"new total-energy : %12g statt %12g\n",e,ea);
#ifdef WITH_PRINTF
     /*    printf("new total-energy : %12g statt %12g\n",e,ea);*/
#endif
    }
   
   if( ((t+1) % TLOG) == 0)
    {
     fprintf(fpl,"%6d) E=%12g , acc=%g\n",t+1,e,100*(double)ac/(double)gc);
#ifdef WITH_PRINTF
     printf("%6d) E=%12g , acc=%g\n",t+1,e,100*(double)ac/(double)gc);
#endif
    }
  }
 e = total_energy();
/* fprintf(fpl,"%6d) E=%12g\n",t,e);*/
#ifdef WITH_PRINTF
/*     printf("%6d) E=%12g , acc=%g\n",t,e,100*(double)ac/(double)gc);*/
#endif

return(0);
}

double diff_energy(int pch)
{
 int p2;
 double dx,dz,dy,dr2,dr6,de,e=0,ea,d;

 for(p2=0 ; p2<PANZ; p2++)
   if(p2!=pch)
    {
     dx = diff_mix(ptx,px[p2]);
     dy = diff_mix(pty,py[p2]);
     dz = diff_miz(ptz,pz[p2]);
     dr2=dx*dx+dy*dy+dz*dz;
     dr6=dr2*dr2*dr2;
     de = 4*(1/dr6/dr6 -1/dr6);
     if(p2<pch)                           /* TRAIL-EINTRAGE in umgekehrter Ordnung*/
      {
       etab[pch][p2] = de;
       e += (de - etab[p2][pch]);
       d = (de - etab[p2][pch]);
       ea = etab[p2][pch];
      }
     else
      {
       etab[p2][pch] = de;
       e += (de - etab[pch][p2]);
       d = (de - etab[pch][p2]);
       ea = etab[pch][p2];
      }
    }
 return(e);
}

double total_energy()
{
 int p1,p2;
 double dx,dz,dy,dr2,dr6,de=0,e=0;

 for(p1=0 ; p1<PANZ; p1++)
   for(p2=p1+1 ; p2<PANZ; p2++)
    {
     dx = diff_mix(px[p1],px[p2]);
     dy = diff_mix(py[p1],py[p2]);
     dz = diff_miz(pz[p1],pz[p2]);
     dr2=dx*dx+dy*dy+dz*dz;
     dr6=dr2*dr2*dr2;
     de = 4*(1/dr6/dr6 -1/dr6);
     e += de;
     etab[p1][p2] = de;
    }
 return(e);
}
/*
double proof_energy(double e_akt)
{
 int p1,p2;
 double dx,dz,dy,dr2,dr6,de,e=0,sd;

 for(p1=0 ; p1<PANZ; p1++)
   for(p2=p1+1 ; p2<PANZ; p2++)
    {
     dx = diff_mix(px[p1],px[p2]);
     dy = diff_mix(py[p1],py[p2]);
     dz = diff_miz(pz[p1],pz[p2]);
     dr2=dx*dx+dy*dy+dz*dz;
     dr6=dr2*dr2*dr2;
     de = 4*(1/dr6/dr6 -1/dr6);
     e += de;
     sd += (de-etab[p1][p2]);
    }
 return(e-e_akt);
}
*/
double diff_mix(double k1,double k2)
{
 double d;

 d=k1-k2;
 if(d>MI_WIDTH)
   d -= WIDTH; 
 if(d<-MI_WIDTH)
   d += WIDTH;
 return(d);
}
double diff_miz(double k1,double k2)
{
 double d;

 d=k1-k2;
 if(d>MI_HEIGHT)
   d-= HEIGHT;
 if(d<-MI_HEIGHT)
   d+= HEIGHT;
 return(d);
}
