
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/times.h>

#include "md_bench.h"

#ifdef SGI64
#define SGI64V2
#endif

void init_rnd(void);
int   rnd(unsigned int *);
double rndf(unsigned int *);
double rndKO(double);
double rndKOS(double);
double gauss(double);

int md_run(int);
void update_pl(void);
void init_structure(void);
double total_energy(void);
double diff_mix(double,double);
double diff_miz(double,double);

double px[PANZ],py[PANZ],pz[PANZ],ptx,pty,ptz;
double vx[PANZ],vy[PANZ],vz[PANZ];
double ax[PANZ],ay[PANZ],az[PANZ];
double dx[PANZ],dy[PANZ],dz[PANZ];
int pairs[PANZ*PANZ/4][2];

double rc2,rl2,dt;
int last_pair;

unsigned int seed,os;

FILE *fpl,*fpt;



/******/
/*MAIN*/
/******/
int main(int argc,char *argv[])
{
 double runTime,runSecs,runMin;
 int trun=-1;
 char file[256];
 struct tms start,stop;

 trun = atoi(argv[1]);

 sprintf(file,"result.%s",argv[2]);

 if(trun<=0)
  {
   printf("MC-Bench needs TIME_STEPS to be specified in mcb.conf\n");
   printf("Terminating\n");
   return(-1);
  }

 printf("running MDBench with %d particles for %d steps\n",PANZ,trun);
 /* printf("%d %d\n",sizeof(unsigned int),sizeof(unsigned long));*/
 init_rnd();
 seed =1234567l;
 fpl = fopen("mdb.log","w");
 if(fpl == NULL)
   printf("Can't open outputfile\n");
 init_structure();

#ifdef WITH_PRINTF
 printf("Starting  MD-bench\n");
#endif

 times(&start);
 md_run(trun);
 times(&stop);

 runTime = ((stop.tms_stime +stop.tms_utime) - (start.tms_stime +start.tms_utime)) / (double)sysconf(_SC_CLK_TCK);
 runSecs = fmod(runTime ,60.);
 runMin  = runTime-runSecs;
 fpt=fopen(file,"w");
 printf("Laufzeit %f:%f\n",runMin,runSecs);
 fprintf(fpt,"%d %f %d\n",trun,runTime,PANZ);
 fclose(fpl);
 fclose(fpt);
 return(0);
}

/***************************************************/
/* Initialisierung der Zufallsgenerator Wortl"ange */
/***************************************************/
void init_rnd() 
{ 
 os   = (unsigned int) 256*256*256*128; 
} 

/***************************************/
/* Erzeugung von integer Zufallszahlen */
/***************************************/
int rnd(unsigned int *seed) 
{ 
 unsigned int j; 
 
 j     = (69069l* *seed+1) ; 
 *seed = j; 
 return(j); 
} 

/***************************************/
/* Erzeugung von double  Zufallszahlen */
/***************************************/
double rndf(unsigned int *seed) 
{ 
 unsigned int j; 
 double r; 
  
 j     = (69069l* *seed+1); 
 *seed = j; 
 r     = (double)j/os*.5; 
 //  printf("seed=%ld -> %f\n",*seed,r);
 return(r); 
} 
/**********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [0,L[ */
/**********************************************************/
double rndKO(double L)
{ 
  //  printf("L=%f\n",L);
 return(L*rndf(&seed));
}

/***********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [-L,L[ */
/***********************************************************/
double rndKOS(double L)
{ 
 return(L*(2*rndf(&seed)-1));
}

/***********************************************************/
/* Erzeugung von double  Zufallszahlen im Intervall [-L,L[ */
/*           mit angen"aherter Gau"sverteilung             */
/***********************************************************/
double gauss(double L)
{
 return(L*((rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)+rndf(&seed)-4)/4));
}



void init_structure()
{
 int n1;
 double ek=0,eks=1.5*PANZ*AKT_TEMP,fv;
 for(n1=0 ;n1<PANZ; n1++)
  {
   px[n1] = rndKO(WIDTH);
   py[n1] = rndKO(WIDTH);
   pz[n1] = rndKO(HEIGHT);
   vx[n1] = gauss(1.0);
   vy[n1] = gauss(1.0);
   vz[n1] = gauss(1.0);
    ek += (vx[n1]*vx[n1] + vy[n1]*vy[n1] + vz[n1]*vz[n1] );
  }
 fv = sqrt(eks/ek);
 for(n1=0;n1<PANZ; n1++)
  {
   vx[n1]  = vx[n1]*fv;
   vy[n1]  = vy[n1]*fv;
   vz[n1]  = vz[n1]*fv;
  }
}
int md_run(int trun)
{
 double dr[3],dr2,dr6,f[3],f0,eks,ek,fv,dr2max=0,dr2_threshold;
 double dt=0.005,e;
 int pn,n1,n2,t,c;
 
 printf("running for %d timesteps (%d %d) \n",trun,TLOG,TNEW);
 eks = 1.5 *PANZ*AKT_TEMP;
 rc2=CUTOFF*CUTOFF;
 rl2=LCUTOFF*LCUTOFF;
 dr2 = (LCUTOFF -CUTOFF)/2;
 dr2_threshold = dr2*dr2;
 update_pl();
 e = total_energy();
 fprintf(fpl,"%6d) E=%12g\n",t,e);
 /*rintf("Start : E=%12g\n",e);*/
 for(t=0 ; t<trun; t++)
  {
   /* Reset Forces/accelerations */
   for(n1=0;n1<PANZ; n1++)
    {
     ax[n1] = 0.0;
     ay[n1] = 0.0;
     az[n1] = 0.0;
    }
   c=0;
   /* Calculated LJ-Force / acceleration for all pairs of particles */
   for(pn=0;pn<last_pair; pn++)
    {
     n1=pairs[pn][0];
     n2=pairs[pn][1];
     dr[0] = px[n1] -px[n2];
     dr[1] = py[n1] -py[n2];
     dr[2] = pz[n1] -pz[n2];
     if(dr[0]>MI_WIDTH)
       dr[0]-=WIDTH;
     if(dr[1]>MI_WIDTH)
       dr[1]-=WIDTH;
     if(dr[2]>MI_HEIGHT)
       dr[2]-=HEIGHT;
     if(dr[0]<-MI_WIDTH)
       dr[0]+= WIDTH;
     if(dr[1]<-MI_WIDTH)
       dr[1]+= WIDTH;
     if(dr[2]<-MI_HEIGHT)
       dr[2]+=HEIGHT;
     
     dr2=dr[0]*dr[0]+dr[1]*dr[1]+dr[2]*dr[2];
     /*   printf("%d/%d) %f<%f ? \n",n1,n2,dr2,rc2);
     printf("r(n1) : %12.8f %12.8f %12.8f\n",px[n1],py[n1],pz[n1]);
     printf("r(n2) : %12.8f %12.8f %12.8f\n",px[n2],py[n2],pz[n2]);
     printf("dr    : %12.8f %12.8f %12.8f\n",dr[0],dr[1],dr[2]);*/
     if(dr2<rc2)
      {
       dr6=dr2*dr2*dr2;
       f0=4*(6/dr6-12/dr6/dr6)/dr2;
       f[0] = f0*dr[0];
       f[1] = f0*dr[1];
       f[2] = f0*dr[2];

       ax[n1] -= f[0];
       ay[n1] -= f[1];
       az[n1] -= f[2];
       ax[n2] += f[0];
       ay[n2] += f[1];
       az[n2] += f[2];
       c++;
       /*     printf("interacting :%d\n",c);*/
      }
    }
 /*     printf("%d force-pairs/ %d active (%d)\n",last_pair,c,PANZ);*/

   ek=0;
   for(n1=0;n1<PANZ; n1++)
    {
     vx[n1]+= ax[n1]*dt;
     vy[n1]+= ay[n1]*dt;
     vz[n1]+= az[n1]*dt;
     ek += (vx[n1]*vx[n1] + vy[n1]*vy[n1] + vz[n1]*vz[n1] );
    }
   fv = sqrt(eks/ek);
   dr2max=0;
   for(n1=0;n1<PANZ; n1++)
    {
     vx[n1]  = vx[n1]*fv;
     vy[n1]  = vy[n1]*fv;
     vz[n1]  = vz[n1]*fv;
     px[n1] += vx[n1]*dt;
     py[n1] += vy[n1]*dt;
     pz[n1] += vz[n1]*dt;
     dx[n1] += vx[n1]*dt;
     dy[n1] += vy[n1]*dt;
     dz[n1] += vz[n1]*dt;
     dr2 = dx[n1]*dx[n1]+dy[n1]*dy[n1]+dz[n1]*dz[n1];
     if(dr2>dr2max)
      {
       dr2max=dr2;
      }
    }
   if(dr2 > dr2_threshold)
    {
     update_pl();
     dr2max=0;
    }
   if( ((t+1) % TLOG) == 0)
    {
     e = total_energy();
     fprintf(fpl,"%6d) E=%12g \n",t+1,e);
#ifdef WITH_PRINTF
     printf("%6d) E=%12g \n",t+1,e);
#endif
    }
   
  }
 
}
void update_pl()
{
 int n1,n2,pn=0;
 double dr[3],dr2;

 for(n1=0; n1<PANZ; n1++)
  {
   dx[n1] = 0.0;
   dy[n1] = 0.0;
   dz[n1] = 0.0;
   for(n2=n1+1; n2<PANZ; n2++)
    {
     dr[0] = px[n1] -px[n2];
     dr[1] = py[n1] -py[n2];
     dr[2] = pz[n1] -pz[n2];
     if(dr[0]>MI_WIDTH)
       dr[0]-=WIDTH;
     if(dr[1]>MI_WIDTH)
       dr[1]-=WIDTH;
     if(dr[2]>MI_HEIGHT)
       dr[2]-=HEIGHT;
     if(dr[0]<-MI_WIDTH)
       dr[0]+= WIDTH;
     if(dr[1]<-MI_WIDTH)
       dr[1]+= WIDTH;
     if(dr[2]<-MI_HEIGHT)
       dr[2]+=HEIGHT;

     dr2=dr[0]*dr[0]+dr[1]*dr[1]+dr[2]*dr[2];
     /*     printf("%d/%d) %f<%f ? \n",n1,n2,dr2,rl2);
     printf("r(n1) : %12.8f %12.8f %12.8f\n",px[n1],py[n1],pz[n1]);
     printf("r(n2) : %12.8f %12.8f %12.8f\n",px[n2],py[n2],pz[n2]);
     printf("dr    : %12.8f %12.8f %12.8f\n",dr[0],dr[1],dr[2]);*/
    if(dr2<rl2)
      {
       pairs[pn][0]=n1;
       pairs[pn][1]=n2;
       pn++;
       /*       printf("PAIR !!! pn=%d\n",pn);*/
      }
    }
  }
 last_pair=pn;
}


double total_energy()
{
 int p1,p2,pn;
 double dx,dz,dy,dr2,dr6,de=0,e=0;
 // printf ("lastpair=%d\n",last_pair);
   for(pn=0;pn<last_pair; pn++)
    {
     p1=pairs[pn][0];
     p2=pairs[pn][1];
     dx = diff_mix(px[p1],px[p2]);
     dy = diff_mix(py[p1],py[p2]);
     dz = diff_miz(pz[p1],pz[p2]);
     dr2=dx*dx+dy*dy+dz*dz;
     if(dr2<rc2)
      {
       dr6=dr2*dr2*dr2;
       de = 4*(1/dr6/dr6 -1/dr6);
       e += de;
      }
     //     printf("\t\t de=%f -> e=%f  %f %f %f    %f %f\n",de,e,pz[p1],pz[p2],dz,dr2,rc2);
     /*     etab[p1][p2] = de;*/
    }
 return(e);
}

double diff_mix(double k1,double k2)
{
 double d;

 d=k1-k2;
 if(d>MI_WIDTH)
   d -= WIDTH; 
 if(d<-MI_WIDTH)
   d += WIDTH;
 return(d);
}
double diff_miz(double k1,double k2)
{
 double d;

 d=k1-k2;
 if(d>MI_HEIGHT)
   d-= HEIGHT;
 if(d<-MI_HEIGHT)
   d+= HEIGHT;
 return(d);
}
